import 'package:bloc/bloc.dart';
import 'package:flutter/widgets.dart';
import 'package:flutter_cubit/flutter_cubit.dart';

import '../flutter_bloc.dart';

/// {@template bloclistener}
/// Takes a [BlocWidgetListener] and an optional [bloc] and invokes
/// the [listener] in response to [state] changes in the [bloc].
/// It should be used for functionality that needs to occur only in response to
/// a [state] change such as navigation, showing a [SnackBar], showing
/// a [Dialog], etc...
/// The [listener] is guaranteed to only be called once for each [state] change
/// unlike the [builder] in [BlocBuilder].
///
/// If the [bloc] parameter is omitted, [BlocListener] will automatically
/// perform a lookup using [BlocProvider] and the current `BuildContext`.
///
/// ```dart
/// BlocListener<BlocA, BlocAState>(
///   listener: (context, state) {
///     // do stuff here based on BlocA's state
///   },
///   child: Container(),
/// )
/// ```
/// Only specify the [bloc] if you wish to provide a [bloc] that is otherwise
/// not accessible via [BlocProvider] and the current `BuildContext`.
///
/// ```dart
/// BlocListener<BlocA, BlocAState>(
///   bloc: blocA,
///   listener: (context, state) {
///     // do stuff here based on BlocA's state
///   },
///   child: Container(),
/// )
/// ```
///
/// An optional [condition] can be implemented for more granular control
/// over when [listener] is called.
/// The [condition] function will be invoked on each [bloc] [state] change.
/// The [condition] takes the previous [state] and current [state] and must
/// return a [bool] which determines whether or not the [listener] function
/// will be invoked.
/// The previous [state] will be initialized to the [state] of the [bloc]
/// when the [BlocListener] is initialized.
/// [condition] is optional and if it isn't implemented, it will default to
/// `true`.
///
/// ```dart
/// BlocListener<BlocA, BlocAState>(
///   condition: (previous, current) {
///     // return true/false to determine whether or not
///     // to invoke listener with state
///   },
///   listener: (context, state) {
///     // do stuff here based on BlocA's state
///   }
///   child: Container(),
/// )
/// ```
/// {@endtemplate}
class BlocListener<B extends Bloc<dynamic, S>, S> extends CubitListener<B, S> {
  /// {@macro bloclistener}
  const BlocListener({
    Key key,
    @required CubitWidgetListener<S> listener,
    B bloc,
    CubitListenerCondition<S> condition,
    Widget child,
  })  : assert(listener != null),
        super(
          key: key,
          child: child,
          listener: listener,
          cubit: bloc,
          condition: condition,
        );
}
